#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <gtk/gtk.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "eggtrayicon.h"
Atom wmAtom;
GMutex * mutx;
EggTrayIcon * mainIcon;
GtkWidget * menu;
GtkTooltips * tips;
gboolean hasMouse=FALSE;
gboolean compizLaunched=FALSE;
gboolean decoratorLaunched=FALSE;
gboolean reloadingCompiz=FALSE;
gboolean reloadingDecorator=FALSE;
gboolean XGL=FALSE;
gboolean useFB=TRUE;
gint WM=0;
gint fallBackWM=0;
gint DM=0;
typedef struct _DMInfo
{
    gchar * Name;
    gchar * Prog;
    GtkWidget * Item;
} DMInfo;
typedef struct _WMInfo
{
    gchar * Name;
    gchar * Prog;
    gchar * Opts;
    gchar * Grep;
    GtkWidget * Item;
    GtkWidget * FBItem;
} WMInfo;
typedef struct _XLSItem
{
    gchar * Command;
    Window w;
} XLSItem;
DMInfo DMs[] = {
    {"Themable Decorator (CGWD)",                "cgwd"                     ,0},
    {"Light Themable Decorator (yawd)",          "yawd"                     ,0},
    {"Gnome Decorator (gnome-window-decorator)", "gnome-window-decorator"   ,0},
};
#define numDM (sizeof(DMs)/sizeof(DMInfo))
// FIXME - commented out WMs are there because they still don't have a decent kill handler.
// this means they don't respect the ICCCM WM_S%d and also aren't listed in xlsclients.
WMInfo WMs[]={
    {"Compiz",                          0,                  0,              0               ,0,0},
    {"Metacity (Gnome Window Manager)", "metacity",         "--replace",    "metacity"      ,0,0},
    {"KWin (KDE Window Manager)",       "kwin",             "--replace",    "kwin"          ,0,0},
//    {"xfwm4 (XFCE Window Manager)",     "xfwm4",            "",             "xfwm4"         ,0,0},
    {"WindowMaker",                     "wmaker",           "",             "WindowMaker"   ,0,0},
//    {"FluxBox",                         "fluxbox",          "",             "fluxbox"       ,0,0},
//    {"BlackBox",                        "blackbox",         "",             "blackbox"      ,0,0},
    {"OpenBox",                         "openbox",          "",             "openbox"       ,0,0},
    {"IceWM",                           "icewm",            "",             "icewm"         ,0,0},
//    {"Enlightenment",                   "enlightenment",    "",             "enlightenment" ,0,0},
};
#define numWM (sizeof(WMs)/sizeof(WMInfo))
GtkWidget * useFBItem;
GtkWidget * FBSubItem;
GtkWidget * DMSubItem;
GtkWidget * reloadDecoratorItem;

void launchWM();

GSList * parse_xlsclients(gchar * output)
{
    GSList * ret = NULL;
    gchar * curpos;
    for(curpos=output;curpos && curpos[0];curpos=strstr(curpos,"\nWindow"))
    {
        XLSItem * i;
        gchar * p;
        gchar * q;
        curpos+=8;
        puts(".");
        i=malloc(sizeof(XLSItem));
        i->w=strtol(curpos,NULL,16);
        p=strstr(curpos,"Command:  ");
        if (!p)
        {
            g_warning("Malformed xlsclients, no command?");
            free(i);
            continue;
        }
        q=strchr(p,'\n');
        if (!q)
        {
            g_warning("Malformed xlsclients, no command?");
            free(i);
            continue;
        }
        q[0]='\0';
        p=g_strdup(p+10);
        q[0]='\n';
        //trim p down to its basename
        q=strchr(p,' ');
        if (q)
            q[0]='\0';
        q=strrchr(p,'/');
        if (!q)
            q=p;
        else
            q++;
        i->Command=g_strdup(q);
        g_free(p);
        ret=g_slist_prepend(ret,i);
    }
    return ret;
}
void load_settings()
{
    gchar * path = g_strconcat(g_get_home_dir(),"/.compiz-managerrc",NULL);
    GKeyFile * f = g_key_file_new();
    if (g_key_file_load_from_file(f,path,0,NULL))
    {
        GError * e = NULL;
        gint i;
        gboolean b;
        i=g_key_file_get_integer(f,"wm-settings","active_wm",&e);
        if (!e)
            WM=i;
        e=NULL;
        i=g_key_file_get_integer(f,"wm-settings","fallback_wm",&e);
        if (!e)
            fallBackWM=i;
        e=NULL;
        i=g_key_file_get_integer(f,"wm-settings","active_dm",&e);
        if (!e)
            DM=i;
        e=NULL;
        b=g_key_file_get_boolean(f,"wm-settings","use_fallback_wm",&e);
        if (!e)
            useFB=b;
        e=NULL;
    }
    g_key_file_free(f);
    g_free(path);
}
void save_settings()
{
    gchar * path = g_strconcat(g_get_home_dir(),"/.compiz-managerrc",NULL);
    GKeyFile * f = g_key_file_new();
    g_key_file_set_integer(f,"wm-settings","active_wm",WM);
    g_key_file_set_integer(f,"wm-settings","fallback_wm",fallBackWM);
    g_key_file_set_integer(f,"wm-settings","active_dm",DM);
    g_key_file_set_boolean(f,"wm-settings","use_fallback_wm",useFB);
    g_file_set_contents(path,g_key_file_to_data(f,NULL,NULL),-1,NULL);
    g_key_file_free(f);
    g_free(path);
}
gboolean is_running(const gchar * command)
{
    gint ex;
    if (!g_spawn_command_line_sync(
                g_strdup_printf("sh -c 'pidof %s > /dev/null'",command),
                NULL,NULL,&ex,NULL))
        g_error("can't use this app, no ps or grep");
    if (WIFEXITED(ex))
    {
        if (WEXITSTATUS(ex)==0)
            return TRUE;
    }
    else
        g_error("something went wrong with ps or grep");
    return FALSE;
}
gboolean is_decor_running()
{
    return (is_running(DMs[DM].Prog));
}

void showMenu(guint button,guint32 time)
{
    gtk_menu_popup(GTK_MENU(menu),NULL,NULL,NULL,NULL,0,time);
}
gboolean buttonUp(GtkWidget * w, GdkEventButton * e, gpointer d)
{
    if (hasMouse)
        showMenu(e->button,e->time);
    return TRUE;
}
gboolean enterNotify(GtkWidget * w, GdkEventCrossing * e, gpointer d)
{
    hasMouse=TRUE;
    return TRUE;
}
gboolean leaveNotify(GtkWidget * w, GdkEventCrossing * e, gpointer d)
{
    hasMouse=FALSE;
    return TRUE;
}
void startApp(GtkWidget * w, gchar * command)
{
    g_spawn_command_line_async(command,NULL);
}
gboolean decoratorSignalled(gint signal)
{
    g_warning("Decorator caught deadly signal %d",signal);
    return FALSE; // terminate for now
}
gboolean compizSignalled(gint signal)
{
    g_warning("Compiz caught deadly signal %d",signal);
    return FALSE; // terminate for now
}
gpointer decorThread(gpointer d)
{
    gint ex=0;
    decoratorLaunched=TRUE;
    while(decoratorLaunched)
    {
        gchar * s = g_strconcat(DMs[DM].Prog," --replace",NULL);
        if (!g_spawn_command_line_sync(s,NULL,NULL,&ex,NULL))
        {
            g_warning("Couldn't launch selected decorator:%s",DMs[DM].Name);
            decoratorLaunched=FALSE;
        }
        else if (reloadingDecorator)
        {
            g_free(s);
            reloadingDecorator=FALSE;
            return NULL;
        }
        else if (WIFEXITED(ex)) // returned something
        {
            if (WEXITSTATUS(ex)) // returned something other than 0
            {
                decoratorLaunched=FALSE;
            }
        }
        else if (WIFSIGNALED(ex)) // killed by a signal
        {
            if (!decoratorSignalled(WTERMSIG(ex)))
            {
                decoratorLaunched=FALSE;
            }
        }
        g_free(s);
    }
    return NULL;
}
void killIfWM(XLSItem * i, Display * d)
{
    gint j;
    puts(i->Command);
    for (j=1;j<numWM;j++)
    {
        if (strcmp(i->Command,WMs[j].Grep)==0)
        {
            g_message("Killing window %x for client %s.",(unsigned int)i->w,WMs[j].Name);
            XKillClient(d,i->w);
        }
    }
}
void killWM()
{
    if (!g_mutex_trylock(mutx))
        return;
    {
        Display * d = GDK_DISPLAY_XDISPLAY(gdk_display_get_default());
        XLockDisplay(d);
        {
            Window w=XGetSelectionOwner(d,wmAtom);
            if (w!=None)
                XKillClient(d,w);
            else
            {
                gchar * cli;
                GSList * xls;
                g_warning("Couldn't find a Selection Owner, perhaps no WM running?\nOtherwise, manually kill your wm, and report the bug to the developers, it doesn't follow the standards.\nFalling back to looking for a defined WM in xlsclients.");
                if (!g_spawn_command_line_sync("xlsclients -l",&cli,NULL,NULL,NULL))
                    g_error("No xlsclients, sorry, can't continue.");
                if (!cli)
                    g_error("No output from xlsclients, bailing.");
                xls=parse_xlsclients(cli);
                g_free(cli);
                g_slist_foreach(xls,(GFunc)killIfWM,d);
            }
            XSync(d,FALSE);
        }
        XUnlockDisplay(d);
    }
    g_mutex_unlock(mutx);
}
void startWM()
{
    gchar * wm_command=g_strjoin(" ",WMs[WM].Prog,WMs[WM].Opts,NULL);
    compizLaunched=FALSE;
    killWM();
    g_spawn_command_line_async(wm_command,NULL);
    g_free(wm_command);
}
gpointer compizThread(gpointer d)
{
    gint ex=0;
    gchar * compiz_command;
    if (XGL)
       compiz_command = "compiz --replace dbus csm";
    else
       compiz_command = 
           "compiz --indirect-rendering --strict-binding --replace dbus csm";
    compizLaunched=TRUE;
    while(compizLaunched)
    {
        killWM();
        g_spawn_command_line_sync(compiz_command,NULL,NULL,&ex,NULL);
        if (reloadingCompiz)
        {
            reloadingCompiz=FALSE;
            return NULL;
        }
        else if (WIFEXITED(ex)) // returned something
        {
            if (WEXITSTATUS(ex)) // returned something other than 0
            {
                compizLaunched=FALSE;
            }
        }
        else if (WIFSIGNALED(ex)) // killed by a signal
        {
            if (!compizSignalled(WTERMSIG(ex)))
            {
                compizLaunched=FALSE;
            }
        }
    }
    //if this thread exits, and WM is still compiz, fall back into the fall back WM
    if (WM==0 && useFB)
    {
        WM=fallBackWM+1;
        gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(WMs[WM].Item),TRUE);
        gtk_widget_set_sensitive(useFBItem,FALSE);
        gtk_widget_set_sensitive(DMSubItem,FALSE);
        gtk_widget_set_sensitive(reloadDecoratorItem,FALSE);
        gtk_widget_set_sensitive(FBSubItem,FALSE);
        for(ex=1;ex<numWM;ex++)
        {
            gtk_widget_set_sensitive(WMs[ex].FBItem,FALSE);
        }
        for(ex=0;ex<numDM;ex++)
        {
            gtk_widget_set_sensitive(DMs[ex].Item,FALSE);
        }
        launchWM();
    }
    return NULL;
}
void startCompiz()
{
    if (!compizLaunched)
        g_thread_create(compizThread,NULL,FALSE,NULL); // launch compiz
    if (!decoratorLaunched && !is_decor_running())
        g_thread_create(decorThread,NULL,FALSE,NULL); // launch decorator
}
void reloadCompiz()
{
    if (compizLaunched)
        reloadingCompiz=TRUE;
    g_thread_create(compizThread,NULL,FALSE,NULL); // launch compiz
    if (!decoratorLaunched && !is_decor_running())
        g_thread_create(decorThread,NULL,FALSE,NULL); // launch decorator
}
void reloadDecorator(GtkWidget * w, gpointer p)
{
    if (decoratorLaunched)
        reloadingDecorator=TRUE;
    g_thread_create(decorThread,NULL,FALSE,NULL); // launch decorator
}
void launchWM()
{
    save_settings();
    if (WM==0)
        startCompiz();
    else
        startWM();
}
void reloadWM(GtkWidget * w, gpointer p)
{
    if (WM==0)
        reloadCompiz();
    else
        startWM();
}
void setWM(GtkWidget * w, gpointer p)
{
    gint i = (gint)p;
    if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(w)))
    {
        gint j;
        if (i!=WM)
        {
            WM=i;
            launchWM();
        }
        gtk_widget_set_sensitive(useFBItem,i==0);
        gtk_widget_set_sensitive(reloadDecoratorItem,i==0);
        gtk_widget_set_sensitive(FBSubItem,i==0 && useFB);
        gtk_widget_set_sensitive(DMSubItem,i==0);
        for(j=0;j<numDM;j++)
        {
            gtk_widget_set_sensitive(DMs[j].Item,i==0);
        }
        for (j=1;j<numWM;j++)
        {
            gtk_widget_set_sensitive(WMs[j].FBItem,i==0 && useFB);
        }
    }
}
void setFallbackWM(GtkWidget * w, gpointer p)
{
    gint i = (gint)p;
    if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(w)))
    {
        fallBackWM=i;
    }
}
void toggleFB(GtkWidget * w, gpointer p)
{
    gint i;
    useFB=gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(w));
    gtk_widget_set_sensitive(FBSubItem,useFB && WM==0);
    for (i=1;i<numWM;i++)
        gtk_widget_set_sensitive(WMs[i].FBItem,useFB && WM==0);
    save_settings();
}
void setDM(GtkWidget * w, gpointer p)
{
    gint i = (gint)p;
    if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(w)))
    {
        DM=i;
        save_settings();
        reloadDecorator(w,p);
    }
}
GtkWidget * make_image(gchar * path)
{
    GdkPixbuf * p;
    gint w,h;
    gtk_icon_size_lookup(GTK_ICON_SIZE_MENU,&w,&h);
    p = gdk_pixbuf_new_from_file_at_size(path,w,h,NULL);
    if (!p)
        return gtk_image_new_from_stock(GTK_STOCK_MISSING_IMAGE,
                    GTK_ICON_SIZE_MENU);
    else
        return gtk_image_new_from_pixbuf(p);
}
void quitMe(GtkWidget * w, gpointer p)
{
    gtk_main_quit();
}
gboolean is_wm_running(gint wm)
{
    const gchar * nam="compiz";
    if (wm>0)
        nam=WMs[wm].Grep;
    return is_running(nam);
}
gboolean detect_app(const gchar * app)
{
    gint ex;
    if (!g_spawn_command_line_sync(
                g_strdup_printf("sh -c 'which %s > /dev/null'",app),
                NULL,NULL,&ex,NULL))
        g_error("can't use this app, no which");
    if (WIFEXITED(ex))
    {
        if (WEXITSTATUS(ex)==0)
            return TRUE;
    }
    else
        g_error("something went wrong with which");
    return FALSE;
}
gboolean detect_wm(gint wm)
{
    if (detect_app(WMs[wm].Prog))
       return TRUE; 
    if (wm == WM)
        WM=-1;
    if (wm == fallBackWM)
       fallBackWM=-1; 
    return FALSE;
}
gboolean detect_dm(gint dm)
{
    if (detect_app(DMs[dm].Prog))
        return TRUE;
    if (dm == DM)
        DM=-1;
    return FALSE;
}
void init_menu()
{
    GtkWidget * mitem;
    GtkWidget * smshell;
    GSList * wms = NULL;
    GSList * fbk = NULL;
    GSList * dms = NULL;
    gint i;
    
    menu = gtk_menu_new();
    g_object_ref(menu);

    mitem = gtk_image_menu_item_new_with_label("Compiz Settings Manager");
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mitem),
            make_image(PIXMAPS_DIR "/csm.png"));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);
    g_signal_connect(mitem,"activate",G_CALLBACK(startApp),"csm");
    
    mitem = gtk_image_menu_item_new_with_label("Compiz Theme Manager");
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mitem),
            make_image(PIXMAPS_DIR "/gcompizthemer-icon.png"));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);
    g_signal_connect(mitem,"activate",G_CALLBACK(startApp),"gcompizthemer");
    
    mitem = gtk_separator_menu_item_new();
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);
    
    mitem = gtk_image_menu_item_new_with_label("Reload Window Manager");
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mitem),
            gtk_image_new_from_stock(GTK_STOCK_REFRESH,GTK_ICON_SIZE_MENU));
    g_signal_connect(mitem,"activate",G_CALLBACK(reloadWM),NULL);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);
    
    mitem = gtk_image_menu_item_new_with_label("Select Window Manager");
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mitem),
            make_image(DATADIR "/compiz/icon.png"));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);
    
    smshell=gtk_menu_new();
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(mitem),smshell);
    
    for (i=0;i<numWM;i++)
    {
        mitem = gtk_radio_menu_item_new_with_label(wms,WMs[i].Name);
        wms=gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(mitem));
        if (i==WM)
            gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mitem),TRUE);
        gtk_menu_shell_append(GTK_MENU_SHELL(smshell),mitem);
        if (i==0 || detect_wm(i))
            gtk_widget_show(mitem);
        else
            gtk_widget_hide(mitem);
        g_signal_connect(mitem,"toggled",G_CALLBACK(setWM),(gpointer)i);
        WMs[i].Item=mitem;
    }

    mitem = gtk_separator_menu_item_new();
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);

    mitem = gtk_image_menu_item_new_with_label("Reload Window Decorator");
    reloadDecoratorItem = mitem;
    gtk_widget_set_sensitive(reloadDecoratorItem,WM==0);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mitem),
            gtk_image_new_from_stock(GTK_STOCK_REFRESH,GTK_ICON_SIZE_MENU));
    g_signal_connect(mitem,"activate",G_CALLBACK(reloadDecorator),NULL);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);

    mitem = gtk_image_menu_item_new_with_label("Select Window Decorator");
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mitem),
            make_image(DATADIR "/compiz/icon.png"));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);

    smshell=gtk_menu_new();
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(mitem),smshell);
    DMSubItem=mitem;

    for (i=0;i<numDM;i++)
    {
        mitem = gtk_radio_menu_item_new_with_label(dms,DMs[i].Name);
        dms=gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(mitem));
        if (i==DM)
            gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mitem),TRUE);
        gtk_menu_shell_append(GTK_MENU_SHELL(smshell),mitem);
        if (detect_dm(i))
            gtk_widget_show(mitem);
        else
            gtk_widget_hide(mitem);
        g_signal_connect(mitem,"toggled",G_CALLBACK(setDM),(gpointer)i);
        DMs[i].Item=mitem;
    }

    mitem = gtk_separator_menu_item_new();
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);

    mitem = gtk_check_menu_item_new_with_label("Launch Fall-back Window\nManager if compiz crashes");
    useFBItem = mitem;
    gtk_widget_set_sensitive(mitem,WM==0);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mitem),useFB);
    g_signal_connect(mitem,"toggled",G_CALLBACK(toggleFB),NULL);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);

    mitem = gtk_image_menu_item_new_with_label("Select Fall-back Window Manager");
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mitem),
            make_image(DATADIR "/compiz/icon.png"));
    FBSubItem=mitem;
    gtk_widget_set_sensitive(FBSubItem,WM==0 && useFB);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);

    smshell=gtk_menu_new();
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(mitem),smshell);

    for (i=1;i<numWM;i++)
    {
        mitem = gtk_radio_menu_item_new_with_label(fbk,WMs[i].Name);
        fbk=gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(mitem));
        if ((i-1)==fallBackWM)
            gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mitem),TRUE);
        gtk_widget_set_sensitive(mitem,WM==0 && useFB);
        gtk_menu_shell_append(GTK_MENU_SHELL(smshell),mitem);
        if (detect_wm(i))
            gtk_widget_show(mitem);
        else
            gtk_widget_hide(mitem);
        WMs[i].FBItem=mitem;
        g_signal_connect(mitem,"toggled",G_CALLBACK(setFallbackWM),(gpointer)i);
    }

    mitem = gtk_separator_menu_item_new();
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);

    mitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT,NULL);
    g_signal_connect(mitem,"activate",G_CALLBACK(quitMe),NULL);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu),mitem);
    gtk_widget_show(mitem);
}
void init_widgets()
{
    GtkWidget * icon;
    GtkWidget * evbox;
    GdkPixbuf * pbuf;
    tips = gtk_tooltips_new();
    mainIcon = egg_tray_icon_new("Compiz Manager");
    evbox = gtk_event_box_new();
    gtk_event_box_set_visible_window(GTK_EVENT_BOX(evbox),FALSE);
    gtk_tooltips_set_tip(tips,evbox,"Compiz Manager","Manage various things related to compiz");
    pbuf = gdk_pixbuf_new_from_file(PIXMAPS_DIR "/compiz-manager.png",NULL);
    if (!pbuf)
        icon = gtk_image_new_from_stock(GTK_STOCK_MISSING_IMAGE,GTK_ICON_SIZE_SMALL_TOOLBAR);
    else
        icon = gtk_image_new_from_pixbuf(pbuf);
    gtk_container_add(GTK_CONTAINER(mainIcon),evbox);
    gtk_container_add(GTK_CONTAINER(evbox),icon);
    //g_signal_connect(evbox,"button-press-event",G_CALLBACK(buttonDown),NULL);
    g_signal_connect(evbox,"button-release-event",G_CALLBACK(buttonUp),NULL);
    g_signal_connect(evbox,"enter-notify-event",G_CALLBACK(enterNotify),NULL);
    g_signal_connect(evbox,"leave-notify-event",G_CALLBACK(leaveNotify),NULL);
    gtk_widget_show_all(GTK_WIDGET(mainIcon));
    init_menu();
}
void detect_xgl()
{
    //find out if we have XGL
    gint ex=0;
    if (!g_spawn_command_line_sync("sh -c 'ps ax | grep -v grep | grep Xgl > /dev/null'",
                NULL,NULL,&ex,NULL))
        g_error("can't use this app, no ps or no grep");
    if (WIFEXITED(ex))
    {
        if (WEXITSTATUS(ex)==0)
            XGL=TRUE;
        else
            XGL=FALSE;
    }
    else
        g_error("something went wrong with ps or grep");
}
int main(int argc, char ** argv)
{
    gint fd;
    gint e;
    Display * d;

    for (fd=0;fd<argc;fd++)
        if (strcmp(argv[fd],"-d")==0)
            break;
    if (fd==argc)
        daemon(0,1);
    
    fd = creat(g_strconcat(g_get_home_dir(),"/.compiz-manager.lock",NULL)
            ,00755);
    e=errno;
    if (fd==-1)
    {
        g_warning("Couldn't open lockfile.\nError:%s",strerror(e));
        return 1;
    }
    if (flock(fd,LOCK_EX | LOCK_NB)!=0)
    {  
        e=errno;      
        g_warning("Can't run more than one compiz-manager.\n(couldn't obtain lock) Error:%s",strerror(e));
        return 2;
    }

    if (!XInitThreads())
    {
        g_warning("Can't init XLib thread support");
        return 3;
    }
    gtk_init(&argc,&argv);

    d = GDK_DISPLAY_XDISPLAY(gdk_display_get_default());
    wmAtom = XInternAtom(d,"WM_S0",0);
    if (!g_thread_supported()) g_thread_init(NULL);
    mutx=g_mutex_new();

    detect_xgl();

    load_settings();
    init_widgets();
    if (WM==-1)
    {
        //set WM to compiz for now
        gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(WMs[0].Item),TRUE);
    }
    if (fallBackWM==-1)
    {
        //we have to find a fall back WM that exists
        gint i;
        for (i=1;i<numWM;i++)
        {
            if (detect_wm(i))
            {
                gtk_check_menu_item_set_active(
                        GTK_CHECK_MENU_ITEM(WMs[i].FBItem),TRUE);
                break;
            }
        }
    }
    if (fallBackWM==-1)
        gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(useFBItem),FALSE);
    if (DM==-1)
    {
        //we have to find a DM that exists
        gint i;
        for (i=0;i<numDM;i++)
        {
            if (detect_dm(i))
            {
                gtk_check_menu_item_set_active(
                        GTK_CHECK_MENU_ITEM(DMs[i].Item),TRUE);
                break;
            }
        }
    }

    if (!is_wm_running(WM))
        launchWM();
    gtk_main();
    return 0;
}
